package infomat.io;

import infomat.vectorspace.*;
import infomat.vectorspace.similarity.*;
import infomat.vectorspace.vector.*;

import mro.util.*;
import mro.util.properties.*;

import java.io.*;
import java.util.*; 
import java.util.zip.*;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * For loading and saving a IObjectGroup. 
 * 
 * @version          070205
 * @author           Magnus Rosell
 */
public class IObjectGroupingIO{

    mro.util.properties.Properties properties;

    private boolean row;
    private IMatrix matrix;
    private IObjectGrouping grouping;

    public IObjectGroupingIO(){ 
	properties = getDefaultProperties();
    }

    public mro.util.properties.Properties getDefaultProperties(){

	mro.util.properties.Properties properties = 
	    new mro.util.properties.Properties("IObjectGrouping Properties", CurrentClassGetter.getClassName());
		
	try{
			
	    Property p = new Property("IObject Identification", 0, "Method for identifying IObjects");
	    properties.addProperty(p);

	    PropertyValueDescription pvd = new PropertyValueDescription("0", "Identification by id number");
	    p.addValueDescription(pvd);
	    pvd = new PropertyValueDescription("1", "Identification by String");
	    p.addValueDescription(pvd);
	    properties.addProperty(p);
	}catch(PropertyNameConflictException pnce){
	    // Won't happen here
	}
	return properties;
    }


    public Object getPropertyValue(String name){
    	return properties.getPropertyValue(name);
    }
    public void setProperty(String name, Object value) throws PropertyNotFoundException{
    	properties.setProperty(name, value);
    }

    public mro.util.properties.Properties getProperties(){
    	return properties;
    }
    public void setProperties(mro.util.properties.Properties properties){
    	this.properties=properties;
    }



    // It would be better if load took an IObjectGroup instead of an IMatrix.
    public IObjectGrouping load(boolean row, IMatrix matrix, File file) {
	this.row=row;
	this.matrix=matrix;
	try {
	    BufferedReader in = 
		new BufferedReader(new FileReader(file));
	    InputSource source = new InputSource(in);
	    return load(source);
	}
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
	return null;
    }

    public IObjectGrouping loadCompressed(boolean row, IMatrix matrix, File file) {
	this.row=row;
	this.matrix=matrix;
	try {
	    // Create a ZipInputStream to read the zip file            
	    ZipInputStream zis = new ZipInputStream(new BufferedInputStream(new FileInputStream(file)));
	    InputSource source = new InputSource(zis);
	    ZipEntry entry;
	    if((entry = zis.getNextEntry()) != null) {
		if(!entry.isDirectory()) {
		    return load(source);
		}
	    }
	    zis.close();
	}
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
	return null;

    }

    private IObjectGrouping load(InputSource source) {
	try {
	    // Create parser+contentHandler and read the XML document
	    XMLReader parser = XMLReaderFactory.createXMLReader();
	    IObjectGroupingHandler handler = new IObjectGroupingHandler(this, row, matrix);
	    parser.setContentHandler(handler);
	    parser.parse(source);

	    IObjectGrouping grouping = handler.getGrouping();
	    if(grouping.numberOfIObjectGroups()>0 && grouping.numberOfIObjects()>0){
		grouping.sort();
		/*
		  System.out.println("********* counted: "+handler.counted()+ "("+grouping.numberOfIObjects()+")");
		  System.out.println("********* groups counted: "+handler.groupsCounted()+"("+grouping.numberOfIObjectGroups()+")");
		*/

		return grouping;
	    }
	    return null;

	    //	    corpus.setUniqueIdentifier(handler.getUniqueIdentifier());
	} catch (Exception e) {
	    System.err.println(e);
	    e.printStackTrace();
	}
	return null;
    }


    /**
     * Exports the Grouping IMatrix in XML format.
     * Uses the SparseSimilarity to export the representation.
     * Only a temporary method.
     */
    public void save(IObjectGrouping grouping, IMatrix matrix, SparseISimilarity sim, File file){
	this.grouping=grouping;
	this.matrix=matrix;
	try {
	    PrintWriter out =
		new PrintWriter(new BufferedWriter(new BufferedWriter(new FileWriter(file))));
	    save(out, sim);
	    out.close();
	}
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
    }


    /**
     * Exports the Grouping IMatrix in XML format
     */
    public void save(IObjectGrouping grouping, IMatrix matrix, File file){
	this.grouping=grouping;
	this.matrix=matrix;
	try {
	    PrintWriter out =
		new PrintWriter(new BufferedWriter(new BufferedWriter(new FileWriter(file))));
	    save(out, null);
	    out.close();
	}
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
    }


    /**
     * Exports the IMatrix in XML format
     * The file is saved in compressed .zip-format.
     */
    public void saveCompressed(IObjectGrouping grouping, IMatrix matrix, String fileName) { 
	this.grouping=grouping;
	this.matrix=matrix;
	try {
	    // Create zip compressed output stream, zip file entry and PrintWriter
	    ZipOutputStream targetStream = new ZipOutputStream(new FileOutputStream(fileName+".xml.zip"));
	    File file = new File(fileName);                        
	    targetStream.putNextEntry(new ZipEntry(file.getName() + ".xml"));
	    PrintWriter out = new PrintWriter(targetStream);
	    save(out, null);            
	    // Close compressed output stream and PrintWriter
	    out.flush();
	    targetStream.closeEntry();
	    targetStream.close();
	    targetStream = null;
	    out.close();
	}
	catch (java.io.IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
    }




    public void printDTD(PrintWriter out) {
	//We use the xml-format specified in corpus.dtd
	//out.println("<?xml version='1.0' encoding='ISO-8859-1'?>");

	out.println("<?xml version='1.0' encoding='ISO-8859-1' standalone=\"yes\"?>");

	out.println("<!DOCTYPE InfomatIObjectGrouping [");

	out.println("<!ELEMENT InfomatIObjectGrouping (Name, IObjectGroups)>");
	out.println("<!ELEMENT Name (#PCDATA)>");
		
	out.println("<!--The IObjectGroups-->");
	out.println("<!ELEMENT IObjectGroups (IObjectGroup+)>");
	out.println("<!ELEMENT IObjectGroup (GroupName, IObjects+)>");

	out.println("<!--The IObjects of one IObjectGroup-->");
	out.println("<!ELEMENT IObjects (IObject+)>");
	out.println("<!ELEMENT IObject (IOString, IOID, IOVal)>");
	out.println("<!ELEMENT IOString (#PCDATA)> <!-- The string representing the IObject -->");
	out.println("<!ELEMENT IOID (#PCDATA)> <!-- IObject id number -->");
	out.println("<!ELEMENT IOVal (#PCDATA)> <!-- Order value -->");

	out.println("]>");
    }



    private void save(PrintWriter out, SparseISimilarity sim) {

	printDTD(out);

	out.println("<InfomatIObjectGrouping>");
	out.println("<Name>"+grouping.getName()+"</Name>");
	out.println("<IObjectGroups>");
	Vector<IObjectGroup> groups = grouping.getGroups();
	for(int i=0; i<groups.size(); i++){
	    IObjectGroup group = groups.elementAt(i);
	    out.println("<IObjectGroup>");
	    out.println("<GroupName>"+group.getName()+"</GroupName>");

	    /////////////////////////
	    if(sim!=null){
		IVector<IVectorIndex> repr = 
		    sim.getGroupRepresentation(group);
		Vector<IVectorIndex> vec = repr.getOrderedVector();
		out.println("<Centroid>");
		for(int j=0; j<vec.size() && j<20; j++){
		    IVectorIndex ii = vec.get(j);
		    IObject io = (IObject) ii;
		    out.println("<IObject>");
		    out.println("<IOString>"+io.getAmbiguousString()+"</IOString>");
		    out.println("<IOID>"+io.getID()+"</IOID>");
		    out.println("<IOVal>"+repr.getValue(ii)+"</IOVal>");				
		    out.println("</IObject>");
		}
		out.println("</Centroid>");
	    }
	    /////////////////////////


	    out.println("<IObjects>");
	    HashMap<IObject, Float> objectOrderValueMap = group.getObjectOrderValueMap();
	    /*
	    Iterator<Map.Entry<IObject,Float>> it = objectOrderValueMap.entrySet().iterator();
	    while(it.hasNext()){
		Map.Entry<IObject,Float> entry = it.next();
		String iostring = entry.getKey().getString();
		float iovalue = entry.getValue();
	    */
	    Vector<IObject> ordered = group.getOrdered();
	    for(int j=0; j<ordered.size(); j++){
		IObject io = ordered.elementAt(j);
		float iovalue = objectOrderValueMap.get(io);

		out.println("<IObject>");
		out.println("<IOString>"+io.getAmbiguousString()+"</IOString>");
		out.println("<IOID>"+io.getID()+"</IOID>");
		out.println("<IOVal>"+iovalue+"</IOVal>");				
		out.println("</IObject>");
	    }
	    out.println("</IObjects>");
	    out.println("</IObjectGroup>");
	}
	out.println("</IObjectGroups>");
	out.println("</InfomatIObjectGrouping>");


	out.flush();
    }




    //    private void saveIObjectSet(IObjectSet set, PrintWriter out){
    /*
      out.println("<!--The Non-Zero Cells of the IMatrix-->");
      out.println("<!ELEMENT Cells (IMatrixCell*)>");
      out.println("<!ELEMENT IMatrixCell (rowID, colID, count?, lw?, w)>");
      out.println("<!ELEMENT rowID (#PCDATA)> <!--Row Object ID-->");
      out.println("<!ELEMENT colID (#PCDATA)> <!--Column Object ID-->");
      out.println("<!ELEMENT c (#PCDATA)> <!--Frequency-->");
      out.println("<!ELEMENT lw (#PCDATA)> <!--Local Weight-->");
      out.println("<!ELEMENT w (#PCDATA)> <!--Weight-->");
    */
    //    }




    private class IObjectGroupingHandler extends DefaultHandler {
	private StringBuffer buffer;

	private int objectIdentification;
	private IMatrix matrix;
	private IObjectSet set;
	private boolean row;
	private IObjectGrouping grouping;
	private HashMap<String, IObject> stringObjectMap;

	private IObjectGroup group;
	private String ioString;
	private int id;
	private float val;
		
	int counter=0;
	int groupCounter =0;
		
	public IObjectGroupingHandler(IObjectGroupingIO iogingIO, boolean row, IMatrix matrix){ //InfomatModel im) {
	    this.matrix =matrix;
	    this.row=row;
	    if(row)
		this.set = matrix.getRowIObjectSet();
	    else				
		this.set = matrix.getColumnIObjectSet();
	    
	    this.objectIdentification = (Integer) iogingIO.getPropertyValue("IObject Identification");

	    if(objectIdentification==1){
		stringObjectMap = new HashMap<String, IObject>();
		Vector<IObject> objects = set.getOrdered();
		for(int i=0; i<objects.size(); i++){
		    IObject io = objects.elementAt(i);
		    stringObjectMap.put(io.getString(), io);
		}
	    }
	}

	public IObjectGrouping getGrouping(){
	    return grouping;
	}
		
	public int counted(){
	    return counter;
	}

	public int groupsCounted(){
	    return groupCounter;
	}

	public void startElement (String namespaceURI, String localName,
				  String qualifiedName, Attributes atts) {
	    buffer = new StringBuffer();

	}

	public void endElement (String namespaceURI, String localName, String qualifiedName) {

	    if(qualifiedName.equals("Name")){
		grouping = new IObjectGrouping(buffer.toString());
	    }else if(qualifiedName.equals("GroupName")){
		group = new IObjectGroup(buffer.toString());
	    }else if(qualifiedName.equals("IObjectGroup")){
		//System.out.println("group size: "+group.numberOfIObjects());
		if(group!=null && group.numberOfIObjects()>0){
		    group.sortInOrderValueOrder();
		    grouping.add(group);
		    groupCounter++;
		    group=null;
		}
	    }

	    else if(qualifiedName.equals("IOString")){
		ioString = buffer.toString();
	    }else if(qualifiedName.equals("IOID")){
		id = Integer.parseInt(buffer.toString());
	    }else if(qualifiedName.equals("IOVal")){
		val = Float.parseFloat(buffer.toString());
	    }else if(qualifiedName.equals("IObject")){
		if(objectIdentification==0){
		    IObject io = set.getIObjectWithId(id);
		    if(io!=null){
			group.add(io, val);
			counter++;
		    }
		}else if(objectIdentification==1){
		    IObject io = stringObjectMap.get(ioString);
		    if(io!=null){
			group.add(io, val);
			counter++;
		    }
		}
	    }			
	}

	public void characters (char ch[], int start, int length) {
	    if(buffer != null)
		buffer.append(ch,start,length);
	}      
    }




}
