package infomat.vectorspace;

import infomat.vectorspace.vector.IVectorIndex;

/**
 * An element in a IMatrix. In many types of applications, as Information
 * Retrieval (IR) for instance, the IMatrix is very sparse. Only the non zero
 * elements are stored. 
 *
 * @param rowObject The "row" IObject of the IMatrixCell.
 * @param columnObject The "column" IObject of the IMatrixCell.
 * @param localCount The number of cooccurences of the "row" and "column" IObject.
 * @param localWeight The local weight of this element/IMatrixCell. It may be calculated in many ways. In IR it is often the term frequency. Here it may be the localCount divided by the sum of all localCounts on the "row" or the "column" of the matrix, depending on which is considered the objects and which the features.
 * @param weight The whole weight of this IMatrixCell. Often the localWeight times the global weight.
 *
 * @version          080128
 * @author           Magnus Rosell
 */
public class IMatrixCell implements IVectorIndex{
    private IObject rowObject;
    private IObject columnObject;

    private float localCount;
    private float localWeight;
    private float weight;

    public IMatrixCell(IObject rowObject, IObject columnObject, 
		       float localCount, float weight){
	this.rowObject=rowObject;
	this.columnObject=columnObject;
	this.localCount=localCount;
	this.weight=weight;
    }

    public IMatrixCell(IObject rowObject, IObject columnObject, 
		       float localCount, float localWeight, float weight){
	this.rowObject=rowObject;
	this.columnObject=columnObject;
	this.localCount=localCount;
	this.localWeight=localWeight;
	this.weight=weight;
    }

	
    public IMatrixCell(IObject rowObject, IObject columnObject, float weight){
	this.rowObject=rowObject;
	this.columnObject=columnObject;
	this.weight=weight;
	this.localCount=-1;
    }

    public IMatrixCell(IMatrixCell mc){	
	this.rowObject=mc.rowObject;
	this.columnObject=mc.columnObject;
	this.localCount=mc.localCount;
	this.localWeight=mc.localWeight;
	this.weight=mc.weight;
    }

    public float getLocalCount(){
	return localCount;
    }

    public float getLocalWeight(){
	return localWeight;
    }

    public float getWeight(){
	return weight;
    }

    public IObject getRowIObject(){
	return rowObject;
    }

    public IObject getColumnIObject(){
	return columnObject;
    }

    public IObject getOtherIObject(IObject io){
	if(io==rowObject)
	    return columnObject;
	else if(io==columnObject)
	    return rowObject;
	return null;
    }

    public void addToLocalCount(float addition){
	localCount += addition;
    }

    public void addOneToLocalCount(){
	localCount++;
    }

    protected void invert(){
	IObject tmp = rowObject;
	rowObject = columnObject;
	columnObject = tmp;
    }
	
    protected void setRowIObject(IObject io){
	rowObject = io;
    }
	
    protected void setColumnIObject(IObject io){
	columnObject = io;
    }

    /**
     * Use with caution!
     */
    public void setLocalWeight(float localWeight){
	this.localWeight=localWeight;
    }

    /**
     * Use with caution!
     */
    public void setWeight(float weight){
	this.weight=weight;
    }

	
    public String getString(){
	return "("+rowObject.getString()+", "+columnObject.getString()+")";
    }

    public String getLocation(){
	return null;
    }

}
