package infomat.vectorspace.similarity;

import infomat.algorithms.grouping.NeedsIMatrix;
import infomat.algorithms.grouping.NeedsIMatrixDimension;
import infomat.vectorspace.*;
import infomat.vectorspace.vector.*;

import mro.util.*;
import mro.util.properties.*;

import java.util.*;

/**
 *
 * @version          080825
 * @author           Magnus Rosell
 */
public class SparseOverlapSimilarity extends SparseISimilarity
    implements NeedsIMatrix, NeedsIMatrixDimension{

    private IMatrix matrix=null;
    private boolean row=true;

    public mro.util.properties.Properties getDefaultProperties(){

	mro.util.properties.Properties properties = 
	    new mro.util.properties.Properties("Overlap Similarity Properties", CurrentClassGetter.getClassName());
		
	try{
	    
	    Property p = new Property("Normalization", 0, "How and if the count is normalized.");
	    PropertyValueDescription pvd = 
		new PropertyValueDescription("0", "None.");
	    p.addValueDescription(pvd);
	    pvd = new PropertyValueDescription("1", "Divided by number of different objects in largest representation.");
	    p.addValueDescription(pvd);
	    pvd = new PropertyValueDescription("2", "Divided by the number of different objects in both representation.");
	    p.addValueDescription(pvd);
	    properties.addProperty(p);

	}catch(PropertyNameConflictException pnce){
	    // Won't happen here
	}
	return properties;
    }

    public String getName(){
    	return "Overlap Similarity";
    }

    public Vector getPropertyVector(){
    	return properties.getPropertyVector();
    }
    

    public Object getPropertyValue(String name){
    	return properties.getPropertyValue(name);
    }

    public void setProperty(String name, Object value) throws PropertyNotFoundException{
    	properties.setProperty(name, value);
    }

    public mro.util.properties.Properties getProperties(){
    	return properties;
    }
    public void setProperties(mro.util.properties.Properties properties){
    	this.properties=properties;
    }


    public SparseOverlapSimilarity(){
	super();
    }

    public SparseOverlapSimilarity(IMatrix matrix, boolean row){
	super();
	this.matrix=matrix;
	this.row=row;
    }

    public SparseOverlapSimilarity(IMatrix matrix, boolean row, 
				   mro.util.properties.Properties properties){
	super();
	this.matrix=matrix;
	this.row=row;
    	this.properties=properties;
    }



    /**
     * Calculates the similarity between two IObject:s.
     * Extracts the representation from the matrix.
     */
    public float sim(IObject io1, IObject io2){

	IVector<IVectorIndex> vec1 = getRepresentation(io1);
	IVector<IVectorIndex> vec2 = getRepresentation(io2);

	return sim(vec1, vec2);
    }

    /**
     * Calculates the similarity between IObject and a IObjectGroup.
     * Extracts the representation from the matrix.
     * Uses the group representation for IObjectGroup:s of all kinds.
     * This takes more time to extract. Therefore it might be better
     * to use the #getGroupRepresentation-method instead...
     */
    public float sim(IObject io, IObjectGroup iog){

	IVector<IVectorIndex> ioVec = getRepresentation(io);
	IVector<IVectorIndex> groupVec = getRepresentation(iog);

	return sim(ioVec, groupVec);
    }

    /**
     * Calculates the similarity between two IObjectGroup:s.
     * Extracts the representation from the matrix.
     * Uses the group representation for IObjectGroup:s of all kinds.
     * This takes more time to extract. Therefore it might be better
     * to use the #getGroupRepresentation-method instead...
     */
    public float sim(IObjectGroup iog1, IObjectGroup iog2){

	IVector<IVectorIndex> vec1 = getGroupRepresentation(iog1);
	IVector<IVectorIndex> vec2 = getGroupRepresentation(iog2);

	return sim(vec1, vec2);
    }


	
    /**
     * Calculates the similarity between two representations.
     *
     * <br><br>
     * Checks whether the second representation is smaller
     * than first. A small decrease in time is achieved by 
     * giving the representation that one believes is smaller 
     * as first parameter.
     * (Do not check it also. That increaes time spent.)
     */

    public <U extends IVectorIndex, V extends IVectorIndex> float sim(IVector<U> vec1, IVector<V> vec2){


	double prod = 0.0f;

	HashMap<U, Float> smallMapA = vec1.getMap();
	HashMap<V, Float> largeMapA = vec2.getMap();
	if(largeMapA.size()<smallMapA.size()){
	    HashMap<V, Float> smallMapB = vec2.getMap();
	    HashMap<U, Float> largeMapB = vec1.getMap();

	    Iterator<Map.Entry<V, Float>> it = 
		smallMapB.entrySet().iterator();
	    while(it.hasNext()){
		Map.Entry<V, Float> entry = it.next();
		Float largeProd = largeMapB.get(entry.getKey());
		if(largeProd!=null){
		    prod += 1;
		}
	    }
	}else{
	    Iterator<Map.Entry<U, Float>> it = 
		smallMapA.entrySet().iterator();
	    while(it.hasNext()){
		Map.Entry<U, Float> entry = it.next();
		Float largeProd = largeMapA.get(entry.getKey());
		if(largeProd!=null)
		    prod += 1;
	    }
	}


	int normalization = (Integer) getPropertyValue("Normalization");
	if(normalization==1){
	    int nom = largeMapA.size();
	    if(nom < smallMapA.size())
		nom = smallMapA.size();
	    prod = prod/nom;
	}else if(normalization==2){
	    HashSet<IVectorIndex> set = new HashSet<IVectorIndex>();
	    set.addAll(smallMapA.keySet());
	    set.addAll(largeMapA.keySet());
	    prod = prod/set.size();
	}



	return (float) prod;
    }



    /**
     * Returns the representation in the form of a IVector<IObject>.
     */
    public IVector<IVectorIndex> getRepresentation(IObject io){

	IVector<IVectorIndex> representation = new IVector<IVectorIndex>();
	HashSet features = null;
	if(getIMatrixDimension())
	    features = getIMatrix().getColumnIMatrixCellsForRowObject(io);
	else
	    features = getIMatrix().getRowIMatrixCellsForColumnObject(io);

	if(features!=null){
	    Iterator it = features.iterator();
	    while(it.hasNext()){
		IMatrixCell sc = (IMatrixCell) it.next();
		IObject feature = null;
		if(getIMatrixDimension())
		    feature = sc.getColumnIObject();
		else
		    feature = sc.getRowIObject();
		representation.put(feature, sc.getWeight());
	    }
	}
	return representation;
    }



	

    public IVector<IVectorIndex> getGroupRepresentation(IObjectGroup iog){
	IVector<IVectorIndex> centroid = new IVector<IVectorIndex>();

	// All IObject:s in the IObjectGroup
	Vector<IObject> ordered = iog.getOrdered();
	for(int i=0; i<ordered.size(); i++){
	    IObject io = ordered.elementAt(i);

	    // All IObjects along the other dimension of the IMatrix
	    HashSet features = null;
	    if(getIMatrixDimension())
		features = getIMatrix().getColumnIMatrixCellsForRowObject(io);
	    else
		features = getIMatrix().getRowIMatrixCellsForColumnObject(io);
	    
	    if(features!=null){
		Iterator it = features.iterator();
		while(it.hasNext()){
		    IMatrixCell sc = (IMatrixCell) it.next();

		    IObject feature = null;
		    if(getIMatrixDimension())
			feature = sc.getColumnIObject();
		    else
			feature = sc.getRowIObject();

		    centroid.add(feature, sc.getWeight());
		}
	    }
	}

	// Average weight
	centroid.multiply(((float) 1.0/iog.numberOfIObjects()));

	return centroid;
    }




    public void setIMatrix(IMatrix matrix){
	this.matrix=matrix;
    }

    public IMatrix getIMatrix(){
	return matrix;
    }

    public void setIMatrixDimension(boolean row){
	this.row=row;
    }

    public boolean getIMatrixDimension(){
	return this.row;
    }

}

