package mro.util.properties;

import mro.util.*;

import java.io.*;
import java.util.*; 

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.SAXException;


/**
 * <p>
 * This class works a bit like java.util.Properties.
 * The main difference is that each value is stored as a 
 * {@link Property}-object,
 * which may have a description attached to it. The Property-objects
 * also may have {@link PropertyValueDescription}:s. These store
 * a description of the meaning of a certain value of the Property.
 * Further the Property:s may be grouped together in 
 * {@link PropertyGroup}:s.
 * </p>
 * <p>
 * Each Properties is tied to a owner object. That object is remembered
 * with a {@link UniqueIdentifier}. The Properties may also have several 
 * related identifiers, also UniqueIdentifiers.
 * </p>
 * <p>
 * When using Properties as a instance field, let the class implement
 * the HasProperties-interface. Set the properties owner to the instance
 * using the {@link #setOwnerIdentifier} method.
 * </p>
 *
 * <p>
 * A Properties-object is conveniently saved/loaded to/from a XML-file.
 * Using the identifiers a datastructure of different objects may be rebuilt.
 * </p>
 *
 * @see Property
 * @see PropertyValueDescription
 * @see PropertyGroup
 * @see mro.util.UniqueIdentifier
 *
 * @version          080220
 * @author           Magnus Rosell
 */

public class Properties{

    /** This identifier identifies the owning object.
	A Properties is not in it self identifiable.
	It should always be tied to one other object.*/
    private UniqueIdentifier ownerIdentifier;

    /** The related identifiers. */
    private Vector<UniqueIdentifier> relatedIdentifiers=null;

    /** The name of the Properties. */
    private String propertiesName=null;

    /** All Property:s. No PropertyGroup:s. */
    private HashMap<String, Property> namePropertyMap = 
	new HashMap<String, Property>();

    /** PropertyGroup:s. Only top-level PropertyGroups. */
    private HashMap<String, PropertyGroup> namePropertyGroupMap = 
	new HashMap<String, PropertyGroup>();

    /** The vector in the order the user wants it.
	Only top-level Property:s and PropertyGroup:s. */
    private Vector<Property> propertyVector = new Vector<Property>();

    /** Info about the Properties (and its owning class). Not used yet.*/
    private String description=null;


    /**
     * A constructor. The other fields need to be set using the methods.
     */
    protected Properties(String name){
	this.propertiesName=name;
    }

    /**
     * A constructor that creates a UniqueIdentifier with the 
     * specified owner class name.
     */
    public Properties(String name, String ownerClassName){
	this.propertiesName=name;
	this.ownerIdentifier=new UniqueIdentifier(ownerClassName);
    }

    /**
     * A constructor that creates a UniqueIdentifier from the object that
     * is passed along and lets it be the owner identifier.
     */
    public Properties(String name, Object o){
	this.propertiesName=name;
	this.ownerIdentifier=new UniqueIdentifier(o);
    }

    /**
     * A constructor. The other fields need to be set using the methods.
     */
    public Properties(String name, UniqueIdentifier ownerIdentifier){
	this.propertiesName=name;
	this.ownerIdentifier=ownerIdentifier;
    }

    /**
     * Copy constructor.
     */
    public Properties(Properties ps) throws PropertyNameConflictException{
	propertiesName=ps.propertiesName;
	ownerIdentifier=ps.ownerIdentifier;

	if(ps.relatedIdentifiers!=null){
	    relatedIdentifiers = new Vector<UniqueIdentifier>();
	    for(int i=0; i<ps.relatedIdentifiers.size(); i++){
		relatedIdentifiers.add(ps.relatedIdentifiers.elementAt(i));
	    }
	}

	// The properties
	for(int i=0; i<ps.propertyVector.size(); i++){
	    Property p = (Property) ps.propertyVector.elementAt(i);

	    if(p instanceof PropertyGroup){
		PropertyGroup pg =(PropertyGroup) p;
		PropertyGroup newPG=new PropertyGroup(pg);
		addProperty(newPG);
	    }else{
		Property newP = new Property(p);
		addProperty(newP);
	    }
	}
    }

    /**
     * Adds a {@link Property} (or a {@link PropertyGroup}) to the Properties.
     */
    public void addProperty(Property p) throws PropertyNameConflictException{
	if(p instanceof PropertyGroup){
	    PropertyGroup pg =(PropertyGroup) p;
	    if(getPropertyGroup(pg.getPropertyGroupName())!=null){
		System.out.println(pg.getPropertyGroupName());
		throw new PropertyNameConflictException();
	    }else{
		namePropertyGroupMap.put(pg.getPropertyGroupName(), pg);
		Vector properties = pg.getPropertyVector();
		recursivelyAddPropertiesInPropertyVector(properties);
	    }
	}else{
	    if(getProperty(p.getName())!=null){
		throw new PropertyNameConflictException();
	    }else{
		namePropertyMap.put(p.getName(), p);
	    }
	}
	propertyVector.add(p);
    }

    /**
     * Used by the {@link #addProperty}-method.
     */
    private void recursivelyAddPropertiesInPropertyVector(Vector properties)
	throws PropertyNameConflictException{

	for(int i=0; i<properties.size(); i++){
	    Property p = (Property) properties.elementAt(i);
	    if(p instanceof PropertyGroup){
		recursivelyAddPropertiesInPropertyVector(((PropertyGroup) p).getPropertyVector());
	    }else{
		if(getProperty(p.getName())!=null){
		    throw new PropertyNameConflictException();
		}else{
		    namePropertyMap.put(p.getName(), p);
		}
	    }
	}
    }


    /**
     * Returns a {@link PropertyGroup} with the specified name.
     * null if it doesn't exist.
     */
    public PropertyGroup getPropertyGroup(String propertyGroupName){
	PropertyGroup pg = namePropertyGroupMap.get(propertyGroupName);
	return pg;
    }

    /**
     * Returns a {@link Property} with the specified name.
     * null if it doesn't exist.
     */
    public Property getProperty(String name){
	Property p = namePropertyMap.get(name);
	return p;
    }

    public Object getPropertyValue(String name){
	return namePropertyMap.get(name).getValue();
    }

    public String getAppropriateValueDescription(String name){
	Property p = namePropertyMap.get(name);
	return p.getAppropriateValueDescription();
    }

    public Vector<Property> getPropertyVector(){
	return propertyVector;
    }

    /**
     * Returns true if all Property:s (including PropertyGroup:s)
     * have the same name and are in the same order, false otherwise. 
     * The values does not have to be the same.
     */
    public boolean sameProperties(Properties properties){
	Vector<Property> propertyVector2 = properties.getPropertyVector();
	if(propertyVector.size() != propertyVector2.size())
	    return false;
	for(int i=0; i<propertyVector.size(); i++){
	    Property p = propertyVector.elementAt(i);
	    Property p2 = propertyVector2.elementAt(i);
	    if(p instanceof PropertyGroup){
		if(!(p2 instanceof PropertyGroup))
		    return false;
		if(!((PropertyGroup) p).sameProperties((PropertyGroup) p2))
		    return false;
	    }else if(p2 instanceof PropertyGroup){
		return false;
	    }else{
		if(p.getName().compareTo(p2.getName()) != 0)
		    return false;
	    }
	}
	return true;
    }


    /**
     * Sets a Property with the given name to the given value.
     * If the Property is not in the Properties 
     * a {@link PropertyNotFoundException} is thrown.
     */
    public void setProperty(String name, Object value) 
	throws PropertyNotFoundException{
	Property p = getProperty(name);
	if(p==null)
	    throw new PropertyNotFoundException();
	p.setValue(value);
    }

    /**
     * Uses #setProperty().
     */
    public void setPropertiesInPropertyVector(Vector<Property> propertyVector)
	throws PropertyNotFoundException{

	for(int i=0; i<propertyVector.size(); i++){
	    Property p = propertyVector.elementAt(i);
	    setProperty(p.getName(), p.getValue());
	}
    }

    /**
     * Removes the Property with the specified name.
     * Use with caution: if a class depends on the Property it may fail.
     * Does not report on wether the Property exists.
     * Uses #removePropertyRecursively().
     */
    public void removeProperty(String propertyName){
	namePropertyMap.remove(propertyName);
	removePropertyRecursively(propertyName, propertyVector);
    }

    private void removePropertyRecursively(String propertyName, 
					   Vector<Property> propertyVector){
	for(int i=0; i<propertyVector.size(); i++){
	    Property p = propertyVector.elementAt(i);
	    if(p instanceof PropertyGroup){
		removePropertyRecursively(propertyName, ((PropertyGroup) p).getPropertyVector());
	    }else{
		if(p.getName().compareTo(propertyName)==0){
		    propertyVector.remove(p);
		    return; // there is only one with that name!
		}
	    }
	}
    }


    /**
     * Returns the owner identifier (a {@link mro.util.UniqueIdentifier}).
     */
    public UniqueIdentifier getOwnerIdentifier(){
	return ownerIdentifier;
    }

    /**
     * Set the owner identifier to the specified 
     * {@link mro.util.UniqueIdentifier}.
     */
    public void setOwnerIdentifier(UniqueIdentifier ownerIdentifier){
	this.ownerIdentifier=ownerIdentifier;
    }

    /**
     * Add a {@link mro.util.UniqueIdentifier} as a related identifier.
     */
    public void addRelatedIdentifier(UniqueIdentifier relatedIdentifier){
	relatedIdentifiers.add(relatedIdentifier);
    }

    /**
     * Returns all related identifiers, a {@link java.util.Vector} 
     * of {@link mro.util.UniqueIdentifier}:s.
     */
    public Vector<UniqueIdentifier> getRelatedIdentifiers(){
	return relatedIdentifiers;
    }

    public String getOwnerClassName(){
	return getOwnerIdentifier().getOwnerClassName();
    }

	
    public String getPropertiesName(){
	return propertiesName;
    }
	
    public int size(){
	return propertyVector.size();
    }
	
    /**
     * Returns true if this and ps has the same owner-object
     * as decided by their owner identifier, 
     * an {@link mro.util.UniqueIdentifier}. False otherwise.
     */
    public boolean sameOwnerIdentifier(Properties ps){
	return ps.equals(this);
    }

    /**
     * Returns true if this and ps has the same related identifiers
     * ({@link mro.util.UniqueIdentifier}-objects). False otherwise.
     * If both Properties has no related identifiers true is returned.
     */
    public boolean sameRelatedIdentifiers(Properties ps){
	if(relatedIdentifiers!=null){
	    if(ps.relatedIdentifiers!=null){
		if(relatedIdentifiers.size() == ps.relatedIdentifiers.size()){
		    for(int i=0; i<relatedIdentifiers.size(); i++){
			if(!relatedIdentifiers.elementAt(i).equals(ps.relatedIdentifiers.elementAt(i)))
			    return false;
		    }
		    return true;
		}
		return false;
	    }
	    return false;
	}
	if(ps.relatedIdentifiers==null)
	    return true;
	return false;
    }


    /********* IO *********/

    private String fixFileName(String name){
	String ret ="";
	String[] split = name.split(" ");
	ret = split[0];
	for(int i=1; i<split.length; i++)
	    ret += "_"+split[i];
	return ret;
    }

    /**
     * Saves a Properties-object to the directory with specified file name.
     */
    public void saveToDirectory(String directoryName) throws IOException{
	String name = getPropertiesName();
	name = fixFileName(name);
	save(new File(directoryName+File.separator+name+".xml"));
    }

    /**
     * Saves a Properties-object to a file with the specified file name.
     */
    public void save(String fileName) throws IOException{
	save(new File(fileName));
    }

    /**
     * Saves a Properties-object to the specified file.
     */
    public void save(File file) throws IOException{
	//	try {
	    File dir = file.getParentFile();
	    if(!dir.exists())
		dir.mkdirs();

	    // XSL
	    String xslFileName = file.getParent()+File.separator+
		"Properties.xsl";

	    PrintWriter out = new PrintWriter(new BufferedWriter(new BufferedWriter(new FileWriter(xslFileName))));
	    printXSL(out);
	    out.close();

	    // XML
	    out = new PrintWriter(new BufferedWriter(new BufferedWriter(new FileWriter(file))));

	    //            printXMLHead(out);
	    printXML(out);
	    out.close();
	    /*
	}
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
	    */
    }

    /**
     * Prints the complete DTD of a Properties-XML-file to a 
     * {@link java.io.PrintWriter}.
     */
    public static void printDTD(PrintWriter pw) throws IOException{
	pw.println("<!DOCTYPE Properties [");
	printPartDTD(pw);

	// UniqueIdentifier
	UniqueIdentifier.staticPrintPartDTD(pw);

	pw.println("]>");
    }

    /**
     * Prints the DTD of a Properties-XML-file to a 
     * {@link java.io.PrintWriter}. Only the elements.
     * This method may be used when the Properties-XML should
     * be part of a larger XML.
     */
    public static void printPartDTD(PrintWriter pw) throws IOException{
	pw.println("<!ELEMENT Properties (PropertiesName, Identfiers, PList)>");
	pw.println("<!--The Name of the Properties-->");
	pw.println("<!ELEMENT PropertiesName (#PCDATA)>");

	pw.println("<!--Identifiers-->");
	pw.println("<!ELEMENT Identifiers (Owner, Related?)>");

	pw.println("<!ELEMENT Owner (UniqueIdentifier)> <!--The owner (an object) of the Properties-->");
	pw.println("<!ELEMENT Related (UniqueIdentifier*)> <!--Related objects-->");

	pw.println("<!--The List of Property:s-->");
	pw.println("<!ELEMENT PList (pg|p)*>");

	pw.println("<!--A Group of Property:s-->");
	pw.println("<!ELEMENT pg (pgn, p*)>");
	pw.println("<!ELEMENT pgn (#PCDATA)> <!--The name of the group-->");

	pw.println("<!--A Property-->");
	pw.println("<!ELEMENT p (n, v, d, pvd*)>");
	pw.println("<!ELEMENT n (#PCDATA)> <!--The name of the Property-->");
	pw.println("<!ELEMENT v (#PCDATA)> <!--The value of the Property-->");
	pw.println("<!ELEMENT d (#PCDATA)> <!--A description of the Property-->");

	pw.println("<!--A Property Value Description-->");
	pw.println("<!ELEMENT pvd (v, d)>");
	pw.println("<!ELEMENT v (#PCDATA)> <!--A value of a Property...-->");
	pw.println("<!ELEMENT d (#PCDATA)> <!--... and its description.-->");
    }

    /**
     * Prints a reference to the Properties-XSL-file to a 
     * {@link java.io.PrintWriter}.
     */
    public static void printPropertiesXSLReference(PrintWriter pw) throws IOException{
	pw.println("<?xml-stylesheet type=\"text/xsl\" href=\"Properties.xsl\"?>");
    }

    /**
     * Prints the Properties to a 
     * {@link java.io.PrintWriter}.
     */
    public void printXML(PrintWriter pw) throws IOException{
	XMLHelper.printXMLHead(pw);
	printPropertiesXSLReference(pw); 
	printDTD(pw);
	printPartXML(pw);
    }

    public void printPartXML(PrintWriter pw) throws IOException{
	pw.println("<Properties>");
	pw.println("<PropertiesName>"+propertiesName+"</PropertiesName>");

	pw.println("<Identifiers>");
	pw.println("<Owner>");
	ownerIdentifier.printPartXML(pw);
	pw.println("</Owner>");

	if(relatedIdentifiers!=null){
	    pw.println("<Related>");
	    for(int i=0; i<relatedIdentifiers.size(); i++)
		relatedIdentifiers.elementAt(i).printPartXML(pw);
	    pw.println("</Related>");
	}
	pw.println("</Identifiers>");

	pw.println("<PList>");
	for(int i=0; i<propertyVector.size(); i++){
	    Property p = propertyVector.elementAt(i);
	    p.printXML(pw);
	}
	pw.println("</PList>");
	pw.println("</Properties>");
    }


    /**
     * Prints a basic XSL-file for Properties XML-files to a 
     * {@link java.io.PrintWriter}.
     */
    public void printXSL(PrintWriter pw) throws IOException{
	pw.println("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>");
	pw.println("<xsl:stylesheet version=\"1.0\"");
	pw.println("xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\">");
	pw.println("<xsl:template match=\"/\">");
	pw.println("<html>");
	pw.println("<body>");
	pw.println("<xsl:apply-templates/>");
	pw.println("</body>");
	pw.println("</html>");
	pw.println("</xsl:template>");
	pw.println("");
	pw.println("<xsl:template match=\"PropertiesName\">");
	pw.println("<h2><xsl:value-of select=\".\"/></h2>");
	pw.println("</xsl:template>");
	pw.println("");
	pw.println("<xsl:template match=\"Identifiers\">");
	pw.println("<h3>Identifiers</h3>");
	pw.println("<b>Owner: </b><xsl:value-of select=\"./Owner/UniqueIdentifier/OwnerClassName\"/><br/>");
	pw.println("<xsl:apply-templates select=\"Related\"/>");
	pw.println("</xsl:template>");
	pw.println("");
	pw.println("<xsl:template match=\"Related\">");
	pw.println("<b>Related: </b>");
	pw.println("<xsl:for-each select=\"./UniqueIdentifier\">");
	pw.println("<xsl:value-of select=\"OwnerClassName\"/>");
	pw.println("<xsl:if test=\"position()!=last()\">");
	pw.println("<xsl:text>, </xsl:text>");
	pw.println("</xsl:if>");
	pw.println("<xsl:if test=\"position()=last()-1\">");
	pw.println("<xsl:text> and </xsl:text>");
	pw.println("</xsl:if>");
	pw.println("</xsl:for-each>");
	pw.println("<br/>");
	pw.println("</xsl:template>");
	pw.println("");
	pw.println("<xsl:template match=\"PList\">");
	pw.println("<h3>Properties</h3>");
	pw.println("<table border=\"1\">");
	pw.println("<xsl:apply-templates select=\"p\"/>");
	pw.println("</table>");
	pw.println("<xsl:apply-templates select=\"pg\"/>");
	pw.println("</xsl:template>");
	pw.println("");
	pw.println("<xsl:template match=\"pg\">");
	pw.println("<p>");
	pw.println("<b><xsl:value-of select=\"pgn\"/></b><br/>");
	pw.println("<table border=\"1\">");
	pw.println("<xsl:apply-templates select=\"p\"/>");
	pw.println("</table>");
	pw.println("</p>");
	pw.println("</xsl:template>");
	pw.println("");
	pw.println("<xsl:template match=\"p\">");
	pw.println("<tr>");
	pw.println("<td><xsl:value-of select=\"n\"/></td><td><xsl:value-of select=\"v\"/></td>");
	pw.println("</tr>");
	pw.println("</xsl:template>");
	pw.println("");
	pw.println("</xsl:stylesheet>");
    }


    /**
     * Loads a Properties-object from a file with the specified file name.
     */
    public static Properties load(String fileName)
	throws IOException, SAXException{
	return load(new File(fileName));
    }


    /**
     * Loads a Properties-object from the specified file.
     */
    public static Properties load(File file)
	throws IOException, SAXException{
	Properties properties = null;
	    BufferedReader in = new BufferedReader(new FileReader(file));
	    InputSource source = new InputSource(in);

	    //	    Properties iop = new Properties("Dummy");
	    properties=loadXML(source);
	    /*
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
	    */
	return properties;
    }



    /**
     * Loads a Properties-object from an {@link java.io.InputSource}.
     */
    public static Properties loadXML(InputSource source) 
	throws IOException, SAXException{
	Properties properties = null;
	//	try {
	    // Create parser+contentHandler and read the XML document
	    XMLReader parser = XMLReaderFactory.createXMLReader();
	    PropertiesHandler handler = new PropertiesHandler();
	    parser.setContentHandler(handler);
	    parser.parse(source);
	    properties = handler.getProperties();
	    if(properties!=null){
		properties.ownerIdentifier=handler.getOwnerIdentifier();
		properties.relatedIdentifiers=handler.getRelatedIdentifiers();
	    }
	    /*
	} catch (Exception e) {
	    System.err.println(e);
	    e.printStackTrace();
	}
	    */
	return properties;
    }


    /****************** OLD *****************/


    public void printEvaluationToWriter(Writer w) throws IOException{
	for(int i=0; i<propertyVector.size(); i++){
	    Property p = propertyVector.elementAt(i);
	    w.write(p.toEvaluationFileString());
	}
    }


}

/**
 * Class used for XML-parsing of {@link Properties}-objects.
 */
class PropertiesHandler extends DefaultHandler {

    StringBuffer buffer;

    Properties properties=null;
    Stack<PropertyGroup> propertyGroupStack = new Stack<PropertyGroup>();

    // For one Property
    String name=null;
    String value=null;
    String description=null;
    boolean string = false;

    // For PropertyValueDescription
    boolean inPVD=false; // true when in a PropertyValueDescription
    Vector<PropertyValueDescription> valueDescriptions=
	new Vector<PropertyValueDescription>();
    String pvdValue=null;
    String pvdDescription=null;

    // Identifiers
    UniqueIdentifier currentIdentifier=new UniqueIdentifier();
    UniqueIdentifier ownerIdentifier=null;
    Vector<UniqueIdentifier> relatedIdentifiers = null;	    
    boolean owner = false;
    boolean related = false;

    public Properties getProperties(){
	return properties;
    }

    public UniqueIdentifier getOwnerIdentifier(){
	return ownerIdentifier;
    }

    public Vector<UniqueIdentifier> getRelatedIdentifiers(){
	return relatedIdentifiers;
    }

    public void startElement (String namespaceURI, String localName,
			      String qualifiedName, Attributes atts) {
	buffer = new StringBuffer();

	if(qualifiedName.equals("pg"))          
	    propertyGroupStack.push(new PropertyGroup());
	else if(qualifiedName.equals("pvd"))
	    inPVD=true;
	else if(qualifiedName.equals("Owner"))
	    owner=true;
	else if(qualifiedName.equals("Related")){
	    relatedIdentifiers = new Vector<UniqueIdentifier>();
	    related=true;
	}
	else if(qualifiedName.equals("str"))
	    string=true;
    }


    public void endElement (String namespaceURI, String localName, String qualifiedName) {

	// UniqueIdentifier
	if(qualifiedName.equals("Owner"))
	    owner=false;
	else if(qualifiedName.equals("Related"))
	    related=false;

	// Save the UniqueIdentifier in "currentIdentifier"
	else if(currentIdentifier.endElementPartXML(buffer, qualifiedName));

	else if(qualifiedName.equals("UniqueIdentifier")){
	    if(owner)
		ownerIdentifier=currentIdentifier;
	    if(related){
		relatedIdentifiers.add(currentIdentifier);
	    }
	    currentIdentifier=new UniqueIdentifier();
	}

	else if(qualifiedName.equals("PropertiesName"))
	    properties = new Properties(buffer.toString());

	else if(qualifiedName.equals("n"))
	    name=buffer.toString();
	else if(qualifiedName.equals("v")){
	    if(inPVD)
		pvdValue=buffer.toString();
	    else
		value=buffer.toString();
	}
	else if(qualifiedName.equals("d")){
	    if(inPVD)
		pvdDescription=buffer.toString();
	    else
		description=buffer.toString();
	}

	// En ny PropertyValueDescription
	else if(qualifiedName.equals("pvd")){
	    inPVD=false;
	    PropertyValueDescription pvd =
		new PropertyValueDescription(pvdValue, pvdDescription);
	    valueDescriptions.add(pvd);
	}

	// En ny Property
	else if(qualifiedName.equals("p")){
	    Property p = new Property(name, value, string, description);
	    if(value.equals("null"))
		p = new Property(name, null, string, description);
	    if(valueDescriptions.size()>0){
		p.addVectorOfValueDescriptions(valueDescriptions);
		valueDescriptions=new Vector<PropertyValueDescription>();
	    }
	    if(propertyGroupStack.size()>0){
		propertyGroupStack.peek().addProperty(p);
	    }else{
		try{
		    properties.addProperty(p);
		}catch(PropertyNameConflictException pe){
		    System.out.println("Conflicting PropertyName: "+p.getName());
		    System.out.println("Aborting!");
		    System.exit(0);
		}
	    }

	    if(string)
		string=false;
	}

	// En PropertyGroup
	else if(qualifiedName.equals("pgn"))
	    propertyGroupStack.peek().setPropertyGroupName(buffer.toString());
	else if(qualifiedName.equals("pg")){
	    PropertyGroup pg = propertyGroupStack.pop();
	    if(propertyGroupStack.size()>0)
		propertyGroupStack.peek().addProperty(pg);
	    else
		try{
		    properties.addProperty(pg);
		}catch(PropertyNameConflictException pe){
		    System.out.println("Conflicting PropertyName: "+pg.getPropertyGroupName());
		    System.out.println("Aborting!");
		    System.exit(0);
		}
	}
    }

    public void characters (char ch[], int start, int length) {
	if(buffer != null)
	    buffer.append(ch,start,length);
    }      
}

